@@ -9,11 +9,11 @@
  *
  *	ASSUMPTION:  sizeof(long) >= sizeof(void*)
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/interfaces/libpgtcl/Attic/pgtclId.c,v 1.29 2001/11/05 17:46:37 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/interfaces/libpgtcl/Attic/pgtclId.c,v 1.37.2.2 2003/02/01 00:22:23 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -144,6 +144,7 @@ static void
 PgWatchProc(ClientData instanceData, int mask)
 {
 }
+
 static int
 PgGetHandleProc(ClientData instanceData, int direction,
 				ClientData *handlePtr)
@@ -193,7 +194,7 @@ PgSetConnectionId(Tcl_Interp *interp, PGconn *conn)
 
 #if TCL_MAJOR_VERSION >= 8
 	connid->notifier_channel = Tcl_MakeTcpClientChannel((ClientData) PQsocket(conn));
-	Tcl_RegisterChannel(interp, connid->notifier_channel);
+	Tcl_RegisterChannel(NULL, connid->notifier_channel);
 #else
 	connid->notifier_socket = -1;
 #endif
@@ -227,6 +228,8 @@ PgGetConnectionId(Tcl_Interp *interp, char *id, Pg_ConnectionId ** connid_p)
 	{
 		Tcl_ResetResult(interp);
 		Tcl_AppendResult(interp, id, " is not a valid postgresql connection", 0);
+		if (connid_p)
+			*connid_p = NULL;
 		return (PGconn *) NULL;
 	}
 
@@ -268,21 +271,45 @@ PgDelConnectionId(DRIVER_DEL_PROTO)
 			 entry = Tcl_NextHashEntry(&hsearch))
 			ckfree((char *) Tcl_GetHashValue(entry));
 		Tcl_DeleteHashTable(&notifies->notify_hash);
-		Tcl_DontCallWhenDeleted(notifies->interp, PgNotifyInterpDelete,
-								(ClientData) notifies);
+		if (notifies->conn_loss_cmd)
+			ckfree((void *) notifies->conn_loss_cmd);
+		if (notifies->interp)
+			Tcl_DontCallWhenDeleted(notifies->interp, PgNotifyInterpDelete,
+									(ClientData) notifies);
 		ckfree((void *) notifies);
 	}
 
 	/*
 	 * Turn off the Tcl event source for this connection, and delete any
-	 * pending notify events.
+	 * pending notify and connection-loss events.
 	 */
-	PgStopNotifyEventSource(connid);
+	PgStopNotifyEventSource(connid, true);
 
 	/* Close the libpq connection too */
 	PQfinish(connid->conn);
 	connid->conn = NULL;
 
+	/*
+	 * Kill the notifier channel, too.	We must not do this until after
+	 * we've closed the libpq connection, because Tcl will try to close
+	 * the socket itself!
+	 *
+	 * XXX Unfortunately, while this works fine if we are closing due to
+	 * explicit pg_disconnect, all Tcl versions through 8.4.1 dump core
+	 * if we try to do it during interpreter shutdown.  Not clear why.
+	 * For now, we kill the channel during pg_disconnect, but during interp
+	 * shutdown we just accept leakage of the (fairly small) amount of memory
+	 * taken for the channel state representation.
+	 * (Note we are not leaking a socket, since libpq closed that already.)
+	 * We tell the difference between pg_disconnect and interpreter shutdown
+	 * by testing for interp != NULL, which is an undocumented but apparently
+	 * safe way to tell.
+	 */
+#if TCL_MAJOR_VERSION >= 8
+	if (connid->notifier_channel != NULL && interp != NULL)
+		Tcl_UnregisterChannel(NULL, connid->notifier_channel);
+#endif
+
 	/*
 	 * We must use Tcl_EventuallyFree because we don't want the connid
 	 * struct to vanish instantly if Pg_Notify_EventProc is active for it.
@@ -316,27 +343,34 @@ PgSetResultId(Tcl_Interp *interp, char *connid_c, PGresult *res)
 		return TCL_ERROR;
 	connid = (Pg_ConnectionId *) Tcl_GetChannelInstanceData(conn_chan);
 
-	for (resid = connid->res_last + 1; resid != connid->res_last; resid++)
+	/* search, starting at slot after the last one used */
+	resid = connid->res_last;
+	for (;;)
 	{
-		if (resid == connid->res_max)
+		/* advance, with wraparound */
+		if (++resid >= connid->res_max)
 			resid = 0;
+		/* this slot empty? */
 		if (!connid->results[resid])
 		{
 			connid->res_last = resid;
-			break;
+			break;				/* success exit */
 		}
+		/* checked all slots? */
+		if (resid == connid->res_last)
+			break;				/* failure exit */
 	}
 
 	if (connid->results[resid])
 	{
-		if (connid->res_max == connid->res_hardmax)
+		/* no free slot found, so try to enlarge array */
+		if (connid->res_max >= connid->res_hardmax)
 		{
 			Tcl_SetResult(interp, "hard limit on result handles reached",
 						  TCL_STATIC);
 			return TCL_ERROR;
 		}
-		connid->res_last = connid->res_max;
-		resid = connid->res_max;
+		connid->res_last = resid = connid->res_max;
 		connid->res_max *= 2;
 		if (connid->res_max > connid->res_hardmax)
 			connid->res_max = connid->res_hardmax;
@@ -495,7 +529,8 @@ error_out:
 typedef struct
 {
 	Tcl_Event	header;			/* Standard Tcl event info */
-	PGnotify	info;			/* Notify name from SQL server */
+	PGnotify   *notify;			/* Notify event from libpq, or NULL */
+	/* We use a NULL notify pointer to denote a connection-loss event */
 	Pg_ConnectionId *connid;	/* Connection for server */
 }	NotifyEvent;
 
@@ -506,7 +541,6 @@ Pg_Notify_EventProc(Tcl_Event *evPtr, int flags)
 {
 	NotifyEvent *event = (NotifyEvent *) evPtr;
 	Pg_TclNotifies *notifies;
-	Tcl_HashEntry *entry;
 	char	   *callback;
 	char	   *svcallback;
 
@@ -516,7 +550,11 @@ Pg_Notify_EventProc(Tcl_Event *evPtr, int flags)
 
 	/* If connection's been closed, just forget the whole thing. */
 	if (event->connid == NULL)
+	{
+		if (event->notify)
+			PQfreeNotify(event->notify);
 		return 1;
+	}
 
 	/*
 	 * Preserve/Release to ensure the connection struct doesn't disappear
@@ -541,17 +579,29 @@ Pg_Notify_EventProc(Tcl_Event *evPtr, int flags)
 		/*
 		 * Find the callback to be executed for this interpreter, if any.
 		 */
-		entry = Tcl_FindHashEntry(&notifies->notify_hash,
-								  event->info.relname);
-		if (entry == NULL)
-			continue;			/* no pg_listen in this interpreter */
-		callback = (char *) Tcl_GetHashValue(entry);
+		if (event->notify)
+		{
+			/* Ordinary NOTIFY event */
+			Tcl_HashEntry *entry;
+
+			entry = Tcl_FindHashEntry(&notifies->notify_hash,
+									  event->notify->relname);
+			if (entry == NULL)
+				continue;		/* no pg_listen in this interpreter */
+			callback = (char *) Tcl_GetHashValue(entry);
+		}
+		else
+		{
+			/* Connection-loss event */
+			callback = notifies->conn_loss_cmd;
+		}
+
 		if (callback == NULL)
-			continue;			/* safety check -- shouldn't happen */
+			continue;			/* nothing to do for this interpreter */
 
 		/*
 		 * We have to copy the callback string in case the user executes a
-		 * new pg_listen during the callback.
+		 * new pg_listen or pg_on_connection_loss during the callback.
 		 */
 		svcallback = (char *) ckalloc((unsigned) (strlen(callback) + 1));
 		strcpy(svcallback, callback);
@@ -562,7 +612,10 @@ Pg_Notify_EventProc(Tcl_Event *evPtr, int flags)
 		Tcl_Preserve((ClientData) interp);
 		if (Tcl_GlobalEval(interp, svcallback) != TCL_OK)
 		{
-			Tcl_AddErrorInfo(interp, "\n    (\"pg_listen\" script)");
+			if (event->notify)
+				Tcl_AddErrorInfo(interp, "\n    (\"pg_listen\" script)");
+			else
+				Tcl_AddErrorInfo(interp, "\n    (\"pg_on_connection_loss\" script)");
 			Tcl_BackgroundError(interp);
 		}
 		Tcl_Release((ClientData) interp);
@@ -578,6 +631,9 @@ Pg_Notify_EventProc(Tcl_Event *evPtr, int flags)
 
 	Tcl_Release((ClientData) event->connid);
 
+	if (event->notify)
+		PQfreeNotify(event->notify);
+
 	return 1;
 }
 
@@ -598,20 +654,46 @@ PgNotifyTransferEvents(Pg_ConnectionId * connid)
 		NotifyEvent *event = (NotifyEvent *) ckalloc(sizeof(NotifyEvent));
 
 		event->header.proc = Pg_Notify_EventProc;
-		event->info = *notify;
+		event->notify = notify;
 		event->connid = connid;
 		Tcl_QueueEvent((Tcl_Event *) event, TCL_QUEUE_TAIL);
-		PQfreeNotify(notify);
 	}
 
 	/*
 	 * This is also a good place to check for unexpected closure of the
 	 * connection (ie, backend crash), in which case we must shut down the
 	 * notify event source to keep Tcl from trying to select() on the now-
-	 * closed socket descriptor.
+	 * closed socket descriptor.  But don't kill on-connection-loss
+	 * events; in fact, register one.
 	 */
 	if (PQsocket(connid->conn) < 0)
-		PgStopNotifyEventSource(connid);
+		PgConnLossTransferEvents(connid);
+}
+
+/*
+ * Handle a connection-loss event
+ */
+void
+PgConnLossTransferEvents(Pg_ConnectionId * connid)
+{
+	if (connid->notifier_running)
+	{
+		/* Put the on-connection-loss event in the Tcl queue */
+		NotifyEvent *event = (NotifyEvent *) ckalloc(sizeof(NotifyEvent));
+
+		event->header.proc = Pg_Notify_EventProc;
+		event->notify = NULL;
+		event->connid = connid;
+		Tcl_QueueEvent((Tcl_Event *) event, TCL_QUEUE_TAIL);
+	}
+
+	/*
+	 * Shut down the notify event source to keep Tcl from trying to
+	 * select() on the now-closed socket descriptor.  And zap any
+	 * unprocessed notify events ... but not, of course, the
+	 * connection-loss event.
+	 */
+	PgStopNotifyEventSource(connid, false);
 }
 
 /*
@@ -633,7 +715,7 @@ PgNotifyInterpDelete(ClientData clientData, Tcl_Interp *interp)
 }
 
 /*
- * Comparison routine for detecting events to be removed by Tcl_DeleteEvents.
+ * Comparison routines for detecting events to be removed by Tcl_DeleteEvents.
  * NB: In (at least) Tcl versions 7.6 through 8.0.3, there is a serious
  * bug in Tcl_DeleteEvents: if there are multiple events on the queue and
  * you tell it to delete the last one, the event list pointers get corrupted,
@@ -649,6 +731,22 @@ NotifyEventDeleteProc(Tcl_Event *evPtr, ClientData clientData)
 {
 	Pg_ConnectionId *connid = (Pg_ConnectionId *) clientData;
 
+	if (evPtr->proc == Pg_Notify_EventProc)
+	{
+		NotifyEvent *event = (NotifyEvent *) evPtr;
+
+		if (event->connid == connid && event->notify != NULL)
+			event->connid = NULL;
+	}
+	return 0;
+}
+
+/* This version deletes on-connection-loss events too */
+static int
+AllNotifyEventDeleteProc(Tcl_Event *evPtr, ClientData clientData)
+{
+	Pg_ConnectionId *connid = (Pg_ConnectionId *) clientData;
+
 	if (evPtr->proc == Pg_Notify_EventProc)
 	{
 		NotifyEvent *event = (NotifyEvent *) evPtr;
@@ -675,10 +773,19 @@ Pg_Notify_FileHandler(ClientData clientData, int mask)
 	 * it internally to libpq; but it will clear the read-ready
 	 * condition).
 	 */
-	PQconsumeInput(connid->conn);
-
-	/* Transfer notify events from libpq to Tcl event queue. */
-	PgNotifyTransferEvents(connid);
+	if (PQconsumeInput(connid->conn))
+	{
+		/* Transfer notify events from libpq to Tcl event queue. */
+		PgNotifyTransferEvents(connid);
+	}
+	else
+	{
+		/*
+		 * If there is no input but we have read-ready, assume this means
+		 * we lost the connection.
+		 */
+		PgConnLossTransferEvents(connid);
+	}
 }
 
 
@@ -686,8 +793,8 @@ Pg_Notify_FileHandler(ClientData clientData, int mask)
  * Start and stop the notify event source for a connection.
  *
  * We do not bother to run the notifier unless at least one pg_listen
- * has been executed on the connection.  Currently, once started the
- * notifier is run until the connection is closed.
+ * or pg_on_connection_loss has been executed on the connection.  Currently,
+ * once started the notifier is run until the connection is closed.
  *
  * FIXME: if PQreset is executed on the underlying PGconn, the active
  * socket number could change.	How and when should we test for this
@@ -708,8 +815,10 @@ PgStartNotifyEventSource(Pg_ConnectionId * connid)
 		if (pqsock >= 0)
 		{
 #if TCL_MAJOR_VERSION >= 8
-			Tcl_CreateChannelHandler(connid->notifier_channel, TCL_READABLE,
-							 Pg_Notify_FileHandler, (ClientData) connid);
+			Tcl_CreateChannelHandler(connid->notifier_channel,
+									 TCL_READABLE,
+									 Pg_Notify_FileHandler,
+									 (ClientData) connid);
 #else
 			/* In Tcl 7.5 and 7.6, we need to gin up a Tcl_File. */
 			Tcl_File	tclfile = Tcl_GetFile((ClientData) pqsock, TCL_UNIX_FD);
@@ -724,14 +833,15 @@ PgStartNotifyEventSource(Pg_ConnectionId * connid)
 }
 
 void
-PgStopNotifyEventSource(Pg_ConnectionId * connid)
+PgStopNotifyEventSource(Pg_ConnectionId * connid, bool allevents)
 {
 	/* Remove the event source */
 	if (connid->notifier_running)
 	{
 #if TCL_MAJOR_VERSION >= 8
 		Tcl_DeleteChannelHandler(connid->notifier_channel,
-							 Pg_Notify_FileHandler, (ClientData) connid);
+								 Pg_Notify_FileHandler,
+								 (ClientData) connid);
 #else
 		/* In Tcl 7.5 and 7.6, we need to gin up a Tcl_File. */
 		Tcl_File	tclfile = Tcl_GetFile((ClientData) connid->notifier_socket,
@@ -742,6 +852,9 @@ PgStopNotifyEventSource(Pg_ConnectionId * connid)
 		connid->notifier_running = 0;
 	}
 
-	/* Kill any queued Tcl events that reference this channel */
-	Tcl_DeleteEvents(NotifyEventDeleteProc, (ClientData) connid);
+	/* Kill queued Tcl events that reference this channel */
+	if (allevents)
+		Tcl_DeleteEvents(AllNotifyEventDeleteProc, (ClientData) connid);
+	else
+		Tcl_DeleteEvents(NotifyEventDeleteProc, (ClientData) connid);
 }
